/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <errno.h>
#include <libgdbus/dbus-system.h>

#include "log.h"
#include "common.h"
#include "dd-display.h"

#define DISPLAY_MAX_BRIGHTNESS  100
#define DISPLAY_MIN_BRIGHTNESS  1
#define DISPLAY_DIM_BRIGHTNESS  0

#define HOLDKEY_BLOCK_BIT		0x1

#define METHOD_SET_REFRESH_RATE	"SetRefreshRate"
#define METHOD_LOCK_STATE		"lockstate"
#define METHOD_UNLOCK_STATE		"unlockstate"
#define METHOD_CHANGE_STATE		"changestate"
#define METHOD_GET_DISPLAY_COUNT	"GetDisplayCount"
#define METHOD_GET_MAX_BRIGHTNESS	"GetMaxBrightness"
#define METHOD_GET_BRIGHTNESS	"GetBrightness"
#define METHOD_SET_BRIGHTNESS	"SetBrightness"
#define METHOD_HOLD_BRIGHTNESS	"HoldBrightness"
#define METHOD_RELEASE_BRIGHTNESS	"ReleaseBrightness"
#define METHOD_GET_ACL_STATUS	"GetAclStatus"
#define METHOD_SET_ACL_STATUS	"SetAclStatus"

#define STR_LCD_OFF   "lcdoff"
#define STR_LCD_DIM   "lcddim"
#define STR_LCD_ON    "lcdon"
#define STR_SUSPEND   "suspend"

#define STR_STAYCURSTATE "staycurstate"
#define STR_GOTOSTATENOW "gotostatenow"

#define STR_HOLDKEYBLOCK "holdkeyblock"
#define STR_NULL         "NULL"

#define STR_SLEEP_MARGIN "sleepmargin"
#define STR_RESET_TIMER  "resettimer"
#define STR_KEEP_TIMER   "keeptimer"

API int display_get_max_brightness(void)
{
	int ret;

	ret = dbus_handle_method_sync(DEVICED_BUS_NAME,
				DEVICED_PATH_DISPLAY,
				DEVICED_INTERFACE_DISPLAY,
				METHOD_GET_MAX_BRIGHTNESS,
				NULL, NULL);
	if (ret < 0)
		return DISPLAY_MAX_BRIGHTNESS;

	_D("get max brightness : %d", ret);
	return ret;
}

API int display_set_brightness_with_setting(int val)
{
	int ret;

	if (val < 0 || val > 100)
		return -EINVAL;

	ret = dbus_handle_method_async_var(DEVICED_BUS_NAME,
					DEVICED_PATH_DISPLAY,
					DEVICED_INTERFACE_DISPLAY,
					METHOD_SET_BRIGHTNESS,
					g_variant_new("(i)", val));
	if (ret < 0)
		_E("no message : failed to setting");

	return ret;
}

static inline char *get_lcd_str(unsigned int val)
{
	switch (val) {
	case LCD_NORMAL:
		return STR_LCD_ON;
	case LCD_DIM:
		return STR_LCD_DIM;
	case LCD_OFF:
		return STR_LCD_OFF;
	case SUSPEND:
		return STR_SUSPEND;
	default:
		return NULL;
	}
}

static void display_change_cb(GVariant *var, void *user_data, GError *err)
{
	int val;

	if (!var)
		return;

	if (!dh_get_param_from_var(var, "(i)", &val)) {
		_E("no message [%s]", g_variant_get_type_string(var));
		g_variant_unref(var);
		return;
	}

	g_variant_unref(var);

	_D("%s-%s : %d", DEVICED_INTERFACE_DISPLAY, METHOD_CHANGE_STATE, val);
}

API int display_change_state(unsigned int s_bits)
{
	char *lcd_str = NULL;
	int ret;

	lcd_str = get_lcd_str(s_bits);
	if (!lcd_str)
		return -EINVAL;

	ret = dbus_handle_method_async_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_DISPLAY,
						DEVICED_INTERFACE_DISPLAY,
						METHOD_CHANGE_STATE,
						g_variant_new("(s)", lcd_str),
						display_change_cb,
						-1,
						NULL);
	if (ret < 0)
		_E("no message : failed to change state");

	_D("%s-%s : %d", DEVICED_INTERFACE_DISPLAY, METHOD_CHANGE_STATE, ret);

	return ret;
}

static void display_lock_cb(GVariant *var, void *user_data, GError *err)
{
	int val;

	if (!var)
		return;

	if (!dh_get_param_from_var(var, "(i)", &val)) {
		_E("no message [%s]", g_variant_get_type_string(var));
		goto out;
	}

	_D("%s-%s : %d", DEVICED_INTERFACE_DISPLAY, METHOD_LOCK_STATE, val);

out:
	g_variant_unref(var);
}

API int display_lock_state(unsigned int s_bits, unsigned int flag,
		      unsigned int timeout)
{
	char *p, *pa[4];
	int ret;

	p = get_lcd_str(s_bits);
	if (!p)
		return -EINVAL;
	pa[0] = p;

	if (flag & GOTO_STATE_NOW)
		/* if the flag is true, go to the locking state directly */
		p = STR_GOTOSTATENOW;
	else
		p = STR_STAYCURSTATE;
	pa[1] = p;

	if (flag & HOLD_KEY_BLOCK)
		p = STR_HOLDKEYBLOCK;
	else
		p = STR_NULL;
	pa[2] = p;

	ret = dbus_handle_method_async_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_DISPLAY,
						DEVICED_INTERFACE_DISPLAY,
						METHOD_LOCK_STATE,
						g_variant_new("(sssi)", pa[0], pa[1], pa[2], timeout),
						display_lock_cb,
						-1,
						NULL);
	if (ret < 0)
		_E("no message : failed to lock state");

	_D("%s-%s : %d", DEVICED_INTERFACE_DISPLAY, METHOD_LOCK_STATE, ret);

	return ret;
}

static void display_unlock_cb(GVariant *var, void *user_data, GError *err)
{
	int val;

	if (!var)
		return;

	if (!dh_get_param_from_var(var, "(i)", &val)) {
		_E("no message [%s]", g_variant_get_type_string(var));
		goto out;
	}

	_D("%s-%s : %d", DEVICED_INTERFACE_DISPLAY, METHOD_UNLOCK_STATE, val);

out:
	g_variant_unref(var);
}

API int display_unlock_state(unsigned int s_bits, unsigned int flag)
{
	char *p, *pa[2];
	int ret;

	p = get_lcd_str(s_bits);
	if (!p)
		return -EINVAL;
	pa[0] = p;

	switch (flag) {
	case PM_SLEEP_MARGIN:
		p = STR_SLEEP_MARGIN;
		break;
	case PM_RESET_TIMER:
		p = STR_RESET_TIMER;
		break;
	case PM_KEEP_TIMER:
		p = STR_KEEP_TIMER;
		break;
	default:
		return -EINVAL;
	}
	pa[1] = p;

	ret = dbus_handle_method_async_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_DISPLAY,
						DEVICED_INTERFACE_DISPLAY,
						METHOD_UNLOCK_STATE,
						g_variant_new("(ss)", pa[0], pa[1]),
						display_unlock_cb,
						-1,
						NULL);
	if (ret < 0)
		_E("no message : failed to unlock state");

	_D("%s-%s : %d", DEVICED_INTERFACE_DISPLAY, METHOD_UNLOCK_STATE, ret);

	return ret;
}
